 /*
 * Project: Device in Jar. Greenhouse.
 * Device:  Auxiliary Atmega16
 * File:    main()
 *
 * License: freeware. Impressed by other cool ideas, will be happy if I inspire to new ideas.
 *
 * © 2010 by Maxim Ushakov; www.maxamuxa.com; head@maxamuxa.com
 */
#include "std.h"

#include "twi.h"
#include "util/delay.h"

//
// Storage for values
//

class StorageForValues : public TWI
{
public:
 StorageForValues();

 //These masks contains zero bit where corresponding pin should be pulled down.
 //All other bits are 1.
 volatile uint8_t m_portB_mask[ 128 ], m_portC_mask[ 128 ], m_portD_mask[ 128 ];

private:
 //overriddens from TWI to operate with real data
 virtual uint16_t readRegister( uint8_t address );
 virtual void writeRegister( uint8_t address, uint16_t value );

 uint16_t m_values[ 3 ][ 8 ]; //they keep information about zeros in the masks
};

StorageForValues::StorageForValues()
{
 uint8_t i, j;
 
 //stored values
 for( i = 0; i < 3; i++ )
   for( j = 0; j < 8; j++ )
     m_values[ i ][ j ] = 0;
 
 //masks should keep pulled up pins...
 for( i = 1; i < 128; i++ )
 {
  m_portB_mask[ i ] = 0xFF;
  m_portC_mask[ i ] = 0xFF;
  m_portD_mask[ i ] = 0xFF;
 }
 
 //... instead of the bits marked by m_values
 m_portB_mask[ 0 ] = 0x00;
 m_portC_mask[ 0 ] = 0x00;
 m_portD_mask[ 0 ] = 0x00;
}

uint16_t StorageForValues::readRegister( uint8_t address )
{
 //top bit is 1 to avoid 0 address
 address &= 0x7F;

 //it is easy, nothing is done to masks
 uint8_t registerNumber = address >> 4, bitNumber = address & 0x0F;
 
 if( registerNumber > 2 || bitNumber > 7 )
  return 0;
 
 return m_values[ registerNumber ][ bitNumber ];
}

void StorageForValues::writeRegister( uint8_t address, uint16_t value )
{
 //top bit is 1 to avoid 0 address
 address &= 0x7F;

 uint8_t registerNumber = address >> 4, bitNumber = address & 0x0F;

 if( registerNumber > 2 || bitNumber > 7 )
  return;

 if( value > 0x7F )
  return;

 volatile uint8_t* selectedArray;
 switch( registerNumber )
 {
  case 0 : selectedArray = m_portB_mask; break;
  case 1 : selectedArray = m_portC_mask; break;
  case 2 : selectedArray = m_portD_mask; break;
  default : return;
 }
 
 uint16_t& storedValue = m_values[ registerNumber ][ bitNumber ];

 //avoit using the algorithm below for the same values
 if( storedValue == value )
  return;

 //first, set the zero (to avoid possible blinking if we remove old zero)
 selectedArray[ value ] ^= _BV( bitNumber );

 //remove old zero
 selectedArray[ storedValue ] |= _BV( bitNumber );

 //change the value
 storedValue = value;
}

//
// Main routines
//

StorageForValues storage;

int main()
{
 //pull-up unused pins to avoid current leakages
 DDRA = 0;
 PORTA = 0xFF;

 //all other pins are outputs
 DDRB = 0xFF;
 DDRC = 0xFC; //leave SCL & SDA for TWI
 DDRD = 0xFF;

 //enable TWI
 storage.enable();

 //initialize things to make PWM
 uint8_t clock = 0;

 while( 1 )
 {
  if( !clock )
  {
   //initial values
   PORTB = storage.m_portB_mask[ 0 ];
   PORTC = storage.m_portC_mask[ 0 ];
   PORTD = storage.m_portD_mask[ 0 ];
  }
  else
  {
   //apply masks
   PORTB &= storage.m_portB_mask[ clock ];
   PORTC &= storage.m_portC_mask[ clock ];
   PORTD &= storage.m_portD_mask[ clock ];
  }

  //shift the time
  clock = ( clock + 1 ) & 0x7F;
 }

 return 0;
}
