/*
 * Project: Device in Jar. Greenhouse.
 * Device:  Auxiliary Atmega16
 * File:    TWI driver for a slave device
 *
 * License: freeware. Impressed by other cool ideas, will be happy if I inspire to new ideas.
 *
 * © 2010 by Maxim Ushakov; www.maxamuxa.com; head@maxamuxa.com
  */
#include "std.h"
#include "twi.h"
#include <avr/interrupt.h> 

//at the moment, every initiation of the TWCR is done by this value
#define TWCR_ORDINARY ( _BV( TWINT ) | _BV( TWEN ) | _BV( TWEA ) | _BV( TWIE ) )

//
// TWI class
//

#define F_TWI 100000UL //speed for TWI

//this pointer is used by the interrupt handler to coomunicate with TWI
static TWI* Twi;

void TWI::enable()
{
 //connect real object
 if( !Twi )
  Twi = this;

 //disable it to set up its state in silence
  TWCR = 0;

  //initialization of the code
  m_address = 0;
  m_state = Waiting;

  //enable internal pull-up resistors
  DDRC &= 0b11111100;
  PORTC |= 0b00000011;

  //initialization of the TWI
  TWBR = uint8_t( F_CPU / F_TWI / 2 - 8 ); //speed
  TWAR = 4; //its address

  //run
  sei(); //nobody warned me that interrupts are disabled at start! I've spent hours...
  TWCR = _BV( TWEA ) | _BV( TWEN ) | _BV( TWIE );
}

void TWI::disable()
{
 TWCR = 0;
}

uint16_t TWI::readRegister( uint8_t address )
{
  //return the inversion
  return ~address;
}

void TWI::writeRegister( uint8_t address, uint16_t value )
{
 //do nothing
}

void TWI::eventArisen( Result eventCode )
{
 //do nothing
}

void TWI::dropState()
{
 m_state = TWI::Waiting;
 TWCR = TWCR_ORDINARY | _BV( TWSTO );
}

//
// Routines for interrupts
//

SIGNAL( TWI_vect )
{
 //only good combinations of state&event are parsed here
 switch( Twi->m_state )
 {
  //either send/receive can come
  case TWI::Waiting :
   switch( TWSR )
   {
    case TW_SR_SLA_ACK :
     Twi->m_state = TWI::ReceiveAddress;
     TWCR = TWCR_ORDINARY;
     return;
    case TW_ST_SLA_ACK :
     Twi->m_state = TWI::SendByte1;

     //send either a value or the address
     if( Twi->m_address )
      Twi->m_value = Twi->readRegister( Twi->m_address );
     else
      Twi->m_value = Twi->m_address;

     TWDR = uint8_t( Twi->m_value );
     TWCR = TWCR_ORDINARY;
     return;
   }
   break;

  case TWI::ReceiveAddress :
   //the address is like another data
   if( TWSR == TW_SR_DATA_ACK )
   {
    Twi->m_state = TWI::ReceiveByte1;
    Twi->m_temporaryAddress = TWDR;
    TWCR = TWCR_ORDINARY;
    return;
   }
   break;
  //first byte: ACK is afforded
  case TWI::ReceiveByte1 :
   if( TWSR == TW_SR_DATA_ACK )
   {
    Twi->m_state = TWI::ReceiveByte2;
    Twi->m_value = TWDR;
    TWCR = TWCR_ORDINARY;
    return;
   }
   break;
  //the same almost as for the first byte
  case TWI::ReceiveByte2 :
   if( TWSR == TW_SR_DATA_ACK )
   {
    Twi->m_state = TWI::ReceiveStop;
    Twi->m_value |= uint16_t( TWDR ) << 8;
    TWCR = TWCR_ORDINARY;
    return;
   }
   break;
  //only the STOP is affordable after 2 bytes received
  case TWI::ReceiveStop :
   if( TWSR == TW_SR_STOP )
   {
    Twi->m_state = TWI::Waiting;

    //the data can be written
    if( Twi->m_temporaryAddress )
     Twi->writeRegister( Twi->m_temporaryAddress, Twi->m_value );
    else
     Twi->m_address = uint8_t( Twi->m_value );

    Twi->eventArisen( TWI::DataReceived );

    TWCR = TWCR_ORDINARY;
    return;
   }
   break;

  //ACK should be received
  case TWI::SendByte1 :
   if( TWSR == TW_ST_DATA_ACK )
   {
    Twi->m_state = TWI::SendByte2;
    TWDR = uint8_t( Twi->m_value >> 8 );
    TWCR = TWCR_ORDINARY;
    return;
   }
   break;
  //ACK should be received as the last byte was sent.
  //I can not explain why.
  case TWI::SendByte2 :
   if( TWSR == TW_ST_DATA_NACK )
   {
    Twi->m_state = TWI::Waiting;
    Twi->eventArisen( TWI::DataSent );
    TWCR = TWCR_ORDINARY;
    return;
   }
   break;
  default :
   break;
 }

 //any bad combination of state&event leads to the fail
 Twi->dropState();
 Twi->eventArisen( TWI::Error );
}