 /*
 * Project: Device in Jar. Greenhouse.
 * Device:  Auxiliary Atmega16
 * File:    TWI driver for a slave device
 *
 * License: freeware. Impressed by other cool ideas, will be happy if I inspire to new ideas.
 *
 * © 2010 by Maxim Ushakov; www.maxamuxa.com; head@maxamuxa.com
 */
/*
  This class is mentioned to support TWI for the slave controller.
  It should:
  1. detect wrong statuses on interrupts and treat them as terminated sessions
  2. return to the initial state if a session is terminated
  3. understand if other side (master) has terminated a session

  Hardware configuration:
  MPU: Atmega16
  Clock: 1MHz
  Own address: 2
  Address of the master: 1

  Other notes:
  1. Zero register points the real address of the requested data
     (if any data is requested, this register should be written firstly).
  2. I do not wonder the way this class is used, so make virtual methods to be overwritten later.
  3. The first object of TWI type is operable (receive events, etc).
*/
#include <util/twi.h>

class TWI
{
public:

  enum Result
  {
    DataReceived, //some data is received successfully
    DataSent, //some data is sent successfully
    Nothing, //nothing special :)
    Error //some error
  };

  //Enable & disable TWI.
  void enable();
  void disable();

  //This method is called to obtain some value.
  //Note:
  // 1. the address is 1 or greater (the zero address is used internally).
  // 2. the default realization returns the inversion of the address field (it is useful when is under debugging).
  virtual uint16_t readRegister( uint8_t address );

  //This method is called when some incoming data should be written to a register.
  virtual void writeRegister( uint8_t address, uint16_t value );

  //this method is called if some event has arisen
  virtual void eventArisen( Result eventCode );

  //there is an experimental system of values to describe states with the hierarchy
  enum State
  {
   Waiting        = 0b00000000,

   Receive        = 0b01000000, //receive
   ReceiveAddress = 0b01010000, //receive an address
   ReceiveData    = 0b01100000, //receive, waiting for the data
   ReceiveByte1   = 0b01100100, //receive, waiting for the first byte
   ReceiveByte2   = 0b01101000, //receive, waiting for the second byte
   ReceiveStop    = 0b01110000, //receive, waiting for the stop

   Send           = 0b10000000, //send
   SendData       = 0b10100000, //some data is sent, waiting for the confirmation
   SendByte1      = 0b10101000, //first byte is sent, waiting for the ACK
   SendByte2      = 0b10110000, //second byte is sent, waiting for the NACK
  } m_state;

  //I hope this method can drop the state from any error
  void dropState();

  uint8_t m_address; //an address written to the 'zero register'
  uint8_t m_temporaryAddress; //it is an address for the current session
  uint16_t m_value; //it is a temporary storage for the current session
};
