 /*
 * Project: Device in Jar. Greenhouse.
 * Device:  Main Atmega32
 * File:    ADC to catch the statuses of buttons & thermo-sensors. It is mentioned to use my hardware either for
 *          the buttons (it lights background IR and compares dark/lighted values)
 *          and for the temperature (it use a format of chosen thermo-sensor to transformate their values)
 *
 * License: freeware. Impressed by other cool ideas, will be happy if I inspire to new ideas.
 *
 * © 2010 by Maxim Ushakov; www.maxamuxa.com; head@maxamuxa.com
 */
#include "adc.h"

#include <avr/io.h>
#include <avr/interrupt.h> 

#include "events.h"
#include "dataStorage.h"

void ButtonsAndADC::initialize()
{
 //enable ADC, enable its interrupt, set the 128 prescaler (125KHz in result)
 ADCSRA = _BV( ADEN ) | _BV( ADIE ) | 0x07;

 //this pin handle the IR leds
 DDRC |= _BV( PIN3 );
}

struct DataForADC
{
 enum State
 {
  stateStop,
  stateDarkADC,
  stateLightADC
 } m_state;

 DataForADC();
 
 //The collected information for each button.
 //For the first run (dark) there are clean values, for the second one (light) - differences.
 uint8_t m_buttonsData[ 5 ];
 
 //The information for the temperature is stored with a smooth filter (last several values are used).
 uint16_t m_temperature[ 2 ];
 
 //during the processing, the index of m_data to fill
 uint8_t m_currentIndex;
};

DataForADC::DataForADC()
{
 //so, we will accumulate values from 0
 m_temperature[ 0 ] = m_temperature[ 1 ] = 0;
}

static DataForADC dataForADC;

//start a conversion, ADC & dataForADC are initialized
static void startADC()
{
 //select a pin
 ADMUX = _BV( ADLAR ) | dataForADC.m_currentIndex;
 
 //run
 ADCSRA |= _BV( ADSC );
}

//process all collected data
static void processData()
{
 uint8_t i;

 //buttons
 for( i = 0; i < 5; i++ )
  buttonsData.m_buttons[ i ] = dataForADC.m_buttonsData[ i ];
 buttonsData.fireEvent_changed();

 //temperature
 uint16_t value1;
 int8_t value2;
 bool changes = false;
 for( i = 0; i < 2; i++ )
 {
  //formula to take the temperature from the TMP36 chip
  value1 = dataForADC.m_temperature[ i ] >> 7; //skip the fraction, get leading empty 7 bits for the next multiply
  value1 = ( value1 * 125 ) >> 8; //here we get the real degrees
  value2 = int8_t( value1 ) - 40; //the result
  
  if( value2 != temperatureData.m_temperature[ i ] )
  {
   changes = true;
   temperatureData.m_temperature[ i ] = value2;
  }
 }

 //update the information
 if( changes )
  temperatureData.fireEvent_changed();
}

//this one receives interrupts from the timer
void adcReceiverFor40ms()
{
 switch( dataForADC.m_state )
 {
  case DataForADC::stateStop:
   dataForADC.m_state = DataForADC::stateDarkADC;
   dataForADC.m_currentIndex = 0;
   startADC();
   break;
  case DataForADC::stateDarkADC:
   if( dataForADC.m_currentIndex == 8 ) //elsewise, the collecting has not finished
   {
    dataForADC.m_state = DataForADC::stateLightADC;
    dataForADC.m_currentIndex = 0;
    startADC();
   }
   break;
  case DataForADC::stateLightADC:
   if( dataForADC.m_currentIndex == 8 )
   {
    processData();
    dataForADC.m_state = DataForADC::stateDarkADC;
    dataForADC.m_currentIndex = 0;
    startADC();
   }
   break;
 }
}

//this one is called when a value is converted
SIGNAL( ADC_vect )
{
 //calculate temperatures (TMP36 from AnalogDevices are used)
 if( dataForADC.m_currentIndex > 4 )
 {
  uint16_t &value = dataForADC.m_currentIndex == 5 ? dataForADC.m_temperature[ 0 ] : dataForADC.m_temperature[ 1 ];
  uint16_t value1, value2;

  //collect 10 bits from the ADC
  value1 = ADCL >> 6;
  value2 = ADCH;
  value1 |= value2 << 2;

  //smooth filter
  value = ( value >> 4 ) * 15; //take 15/16 from the old value
  value += value1 << 3; //add the 1/16 of the new value (note: values are kept with 7 fraction bits)
 }
 //calculate differences between light/dark values
 else
 {
 //links to the current button
  uint8_t& data = dataForADC.m_buttonsData[ dataForADC.m_currentIndex ];

  if( dataForADC.m_state == DataForADC::stateDarkADC )
   data = ADCH;
  else
  {
   if( data < ADCH )
    data = ADCH - data;
   else
    data = 0; //to avoid possible underflows
  }
 }

 if( ++dataForADC.m_currentIndex < 8 )
  startADC();
 else
 //change the highlight right now so its transient state will finished in time
 {
  if( dataForADC.m_state == DataForADC::stateDarkADC )
   PORTC |= _BV( PIN3 ); //enable IR leds
  else
   PORTC &= ~_BV( PIN3 ); //disable
 }
}
