 /*
 * Project: Device in Jar. Greenhouse.
 * Device:  Main Atmega32
 * File:    One of the bases of the idea. Data storages keep every data interesting for other beings of the system.
 *          Each data storage send events to mark it has been changed. Other subsystems react on that.
 *
 * License: freeware. Impressed by other cool ideas, will be happy if I inspire to new ideas.
 *
 * © 2010 by Maxim Ushakov; www.maxamuxa.com; head@maxamuxa.com
 */
#ifndef __DATA_STORAGE_H__
#define __DATA_STORAGE_H__

#include <avr/io.h>
#include "events.h"

struct DataStorage
{
 //This method should be called when a data is changed.
 //It fires specific notifications to subscribbed subsystems.
 virtual void fireEvent_changed() = 0; 
};

struct Time
{
 Time( uint8_t hour, uint8_t minute, uint8_t second );

 //Fields are accessible directly.
 //Note : m_hour can be bigger than 24 (if we calculate shifts to compare with dawn/sunsets), it is OK.
 uint8_t m_hour, m_minute, m_second;

 //these methods allow to check dawns & sunsets
 bool operator >( Time const& secondOperand ) const;
};

//Current time.
struct ClockData : public DataStorage
{
 Time m_time;
 uint8_t m_dayOfWeek; //Monday == 0
 
 bool isWeekday() const;
 
 ClockData();
 virtual void fireEvent_changed();
};

//Data coming from the ADC, IR-sensors.
struct ButtonsData : public DataStorage
{
 //these values come from the ADC
 uint8_t m_buttons[ 5 ];

 virtual void fireEvent_changed();
};

//Data coming from the ADC, thermo-sensors.
struct TemperatureData : public DataStorage
{
 //these values come from the ADC
 int8_t m_temperature[ 2 ]; //degrees of Celcium
 
 //checks if a value is correct for the used thermo-sensors
 static bool isValid( int8_t temperature );
 
 virtual void fireEvent_changed();
};

//Stored data defining the climate.
struct ClimateData : public DataStorage
{
 uint8_t m_nightTemperature, m_dayTemperature;
 uint8_t m_dayLength;
 uint8_t m_wateringInterval; //number of days

 ClimateData();
 virtual void fireEvent_changed();
};

//Something that could be called 'state': everything calculated, based on other information.
struct StateData : public DataStorage
{
 //around climate
 bool m_dayNow; //depends on the clock time & the climate

 //bad news
 bool m_watering, m_badTemperature;
 
 //set to true if a heater is on
 bool m_heaterWorks[ 2 ];
 
 StateData();
 virtual void fireEvent_changed();
};

//The heap that keeps all unsorted data.
struct HeapData : public DataStorage
{
 //light alarm for weekends/weekdays
 Time m_weekdayDawn, m_weekendDawn;
 
 //interval for the manual light
 Time m_manualLightInterval;
 
 HeapData();
 virtual void fireEvent_changed();
};

//
// The only data storages with global scope
//
extern ClockData clockData;
extern ButtonsData buttonsData;
extern TemperatureData temperatureData;
extern ClimateData climateData;
extern StateData stateData;
extern HeapData heapData;

#endif //__DATA_STORAGE_H__