/*
 * Project: Device in Jar. Greenhouse.
 * Device:  Main Atmega32
 * File:    One of the bases of the idea. Subsystems represent small 'individual' beings of the system that
 *          receive some events, think something, change some data and send other events
 *
 * License: freeware. Impressed by other cool ideas, will be happy if I inspire to new ideas.
 *
 * © 2010 by Maxim Ushakov; www.maxamuxa.com; head@maxamuxa.com
 */
#include "MPUSettings.h"

#include "subSystem.h"
#include "dataStorage.h"
#include "lcd.h"

//
// ClockSubsystem
//
void ClockSubsystem::eventsHandler( Events event )
{
 //it is the only known event
 if( event != event_SystemTime_second )
  return;
 
 Time& time = clockData.m_time;
 
 time.m_second++;
 if( time.m_second == 60 )
 {
  time.m_second = 0;
  time.m_minute++;
  if( time.m_minute == 60 )
  {
   time.m_minute = 0;
   time.m_hour++;
   if( time.m_hour == 24 )
   {
    time.m_hour = 0;
    clockData.m_dayOfWeek++;
    if( clockData.m_dayOfWeek == 7 )
     clockData.m_dayOfWeek = 0;
   }
   throwEvent( event_ClockSubsystem_hourElapsed );
  }
  throwEvent( event_ClockSubsystem_minuteElapsed );
 }
 clockData.fireEvent_changed();
}

//
// LCDSubsystem
//
LCDSubsystem::LCDSubsystem()
 : m_state( stateDefault ), m_screen( 0 ), m_cursor( 0 )
{
}

void LCDSubsystem::eventsHandler( Events event )
{
 switch( event )
 {
  //
  //timer to blink editable fields
  //
  case event_SystemTime_40ms:
  {
    static uint8_t counter;
    counter++;
    if( counter == 26 )
     counter = 0;
    
    bool fieldShown;
    if( m_state == stateEdit )
    {
      switch( counter )
      {
       case 0:
        fieldShown = true;
        break;
       case 13:
        fieldShown = false;
        break;
       default:
        return;
      }

      FieldDescriptor& field = getCurrentFields()[ m_cursor ];
      if( fieldShown )
       showField( field );
      else
      {
       uint8_t
         currentWidth,
         currentY = field.m_y, highY = currentY + field.m_height;
       for( ; currentY < highY; currentY++ )
       {
        lcd_gotoxy( field.m_x, currentY );
        for( currentWidth = 0; currentWidth < field.m_width; currentWidth++ )
         lcd_putc( ' ' );
       }
      }
    }
  }

  //
  //changes of the interesting data
  //

  case event_ClockData_changed:
  case event_TemperatureData_changed:
  case event_ClimateData_changed:
  case event_StateData_changed:
  case event_HeapData_changed:
  {
   FieldDescriptor* fields = getCurrentFields();
   uint8_t i;
   for( i = 0; fields[ i ].m_id != fieldLastField; i++ )
    if( fields[ i ].m_eventToUpdate == event )
     showField( fields[ i ] );
  }
  return;

  //
  //the Buttons
  //
   
  //go to the first screen, anyway
  case event_ButtonsSubsystem_userLeaves:
   m_screen = 0;
   setState( stateDefault );
   return;
  
  //move up/down or right/left
  case event_ButtonsSubsystem_shiftUp:
    switch( m_state )
    {
    case stateDefault:
      if( m_screen > 0 )
      {
        m_screen--;
        refreshScreen();
      }
      return;
    case stateEdit:
    {
     uint8_t newCursor = m_cursor + 1;
     if( findEditableDown( newCursor ) )
     {
      m_cursor = newCursor;
      refreshScreen();
     }
    }
    return;
    }
    return;
  case event_ButtonsSubsystem_shiftDown:
   switch( m_state )
   {
    case stateDefault:
     if( m_screen < 3 )
     {
      m_screen++;
      refreshScreen();
     }
     return;
    case stateEdit:
     if( m_cursor > 0 )
     {
      uint8_t newCursor = m_cursor - 1;
      if( findEditableUp( newCursor ) )
      {
       m_cursor = newCursor;
       refreshScreen();
      }
     }
     return;
   }
   return;

  //enable/disable the edit mode
  case event_ButtonsSubsystem_switchMode:
   setState( m_state == stateDefault ? stateEdit : stateDefault );
   return;
   
  //change values
  case event_ButtonsSubsystem_inputUp:
   if( m_state == stateEdit )
    editField( true, false );
   return;
  case event_ButtonsSubsystem_inputDown:
   if( m_state == stateEdit )
    editField( false, false );
   return;
  case event_ButtonsSubsystem_inputRepeatedUp:
   if( m_state == stateEdit )
    editField( true, true );
   return;
  case event_ButtonsSubsystem_inputRepeatedDown:
   if( m_state == stateEdit )
    editField( false, true );
   return;
 }
}

void LCDSubsystem::showField( FieldDescriptor& field )
{
 //almost all fields use their only string, let us set it
 lcd_gotoxy( field.m_x, field.m_y );
 
 switch( field.m_id )
 {
  //current time
  case fieldTime_hour :
   lcd_puti_2d_z( clockData.m_time.m_hour );
   break;
  case fieldTime_minute :
   lcd_puti_2d_z( clockData.m_time.m_minute );
   break;
  case fieldDayOfWeek :
   lcd_putc( clockData.m_dayOfWeek > 4 ? LCDSubsystem::symbolWeekend : LCDSubsystem::symbolWeekday );
   lcd_putc( ' ' );
   switch( clockData.m_dayOfWeek )
   {
    case 0: lcd_puts( "Mo" ); break;
    case 1: lcd_puts( "Tu" ); break;
    case 2: lcd_puts( "We" ); break;
    case 3: lcd_puts( "Th" ); break;
    case 4: lcd_puts( "Fr" ); break;
    case 5: lcd_puts( "Sa" ); break;
    case 6: lcd_puts( "Su" ); break;
   }
   break;
   
  //state
  case fieldState :
  {
   lcd_gotoxy( 6, 0 );
   lcd_puts( "      " );
   lcd_gotoxy( 6, 1 );
   lcd_puts( "      " );
   
   //about danger events
   bool danger = false;
   if( stateData.m_watering )
   {
    lcd_gotoxy( 8, 0 );
    lcd_putc(LCDSubsystem::symbolWatering );
    danger = true;
   }
   if( stateData.m_badTemperature )
   {
    lcd_gotoxy( 7, 1 );
    lcd_putc( 0xDC ); //right arrow
    lcd_putc( LCDSubsystem::symbolCelcium );
    lcd_putc( 0xDB ); //left arrow
    danger = true;
   }
   if( !danger )
   {
    lcd_gotoxy( 7, 0 );
    lcd_puts( "I'm" );
    lcd_gotoxy( 6, 1 );
    lcd_puts( "happy" );
   }
   
   //about the heater
   lcd_gotoxy( 15, 0 );
   lcd_putc( stateData.m_heaterWorks[ 0 ] ? 0xD9 : 0xDA );
   lcd_gotoxy( 15, 1 );
   lcd_putc( stateData.m_heaterWorks[ 1 ] ? 0xD9 : 0xDA );
   
   //about the climate
   lcd_gotoxy( 0, 1 );
   lcd_putc( stateData.m_dayNow ? LCDSubsystem::symbolDay : LCDSubsystem::symbolNight );
   lcd_putc( ' ' );
   lcd_puti_2d_s( stateData.m_dayNow ? climateData.m_dayTemperature : climateData.m_nightTemperature );
  }
  break;
   
  //temperature
  case fieldCurrentTemperature_1 :
   showTemperature( temperatureData.m_temperature[ 0 ] );
   break;
  case fieldCurrentTemperature_2 :
   showTemperature( temperatureData.m_temperature[ 1 ] );
   break;

 //climate
 case fieldClimateDayLength :
  lcd_puti_2d_s( climateData.m_dayLength );
  lcd_gotoxy( field.m_x, field.m_y + 1 );
  lcd_puti_2d_s( 24 - climateData.m_dayLength );
  break;
 case fieldClimateDayTemperature :
  lcd_puti_2d_s( climateData.m_dayTemperature );
  break;
 case fieldClimateNightTemperature :
  lcd_puti_2d_s( climateData.m_nightTemperature );
  break;
 case fieldClimateWateringInterval :
  lcd_puti_2d_s( climateData.m_wateringInterval );
  break;
 
 //morning alarms
 case fieldDawnTimeWeekday_hour :
  lcd_puti_2d_z( heapData.m_weekdayDawn.m_hour );
  break;
 case fieldDawnTimeWeekday_minute :
  lcd_puti_2d_z( heapData.m_weekdayDawn.m_minute );
  break;
 case fieldDawnTimeWeekend_hour :
  lcd_puti_2d_z( heapData.m_weekendDawn.m_hour );
  break;
 case fieldDawnTimeWeekend_minute :
  lcd_puti_2d_z( heapData.m_weekendDawn.m_minute );
  break;
  
 //manual time
 case fieldManualLampTime_hour :
  lcd_puti_2d_z( heapData.m_manualLightInterval.m_hour );
  break;
 case fieldManualLampTime_minute :
  lcd_puti_2d_z( heapData.m_manualLightInterval.m_minute );
  break;
 }
}

uint8_t LCDSubsystem::m_customSymbols[]=
{
 0b00100, 0b00100, 0b01010, 0b10111, 0b10111, 0b11111, 0b01110, 0b00000, 
 0b01010, 0b00100, 0b01110, 0b10001, 0b11101, 0b10001, 0b01110, 0b00000,
 0b10000, 0b00110, 0b01001, 0b01000, 0b01001, 0b00110, 0b00000, 0b00000,
 0b00000, 0b10101, 0b01010, 0b10001, 0b01010, 0b10101, 0b00000, 0b00000,
 0b00000, 0b01000, 0b10000, 0b11001, 0b11111, 0b01110, 0b00000, 0b00000,
 0b01010, 0b01111, 0b01110, 0b01010, 0b00010, 0b00010, 0b00010, 0b00000,
 0b00110, 0b00111, 0b00110, 0b10010, 0b11110, 0b11110, 0b01100, 0b00000,
 0b11111, 0b01110, 0b01110, 0b00100, 0b01010, 0b01010, 0b11111, 0b00000
};

void LCDSubsystem::initialize()
{
 lcd_command( _BV( LCD_CGRAM ) );
 uint8_t i;
 for( i = 0; i < 64; i++ )
 {
  lcd_data( m_customSymbols[ i ] );
 }

 refreshScreen();
}

uint8_t LCDSubsystem::m_blankScreens[ 4 ][ 32 ] = 
{
  {
    ' ', ' ', ':', ' ', ' ', ' ', ' ', ' ',
    ' ', ' ', ' ', ' ', ' ', ' ', LCDSubsystem::symbolCelcium, ' ',
    ' ', ' ', ' ', ' ', LCDSubsystem::symbolCelcium, ' ', ' ', ' ',
    ' ', ' ', ' ', ' ', ' ', ' ', LCDSubsystem::symbolCelcium, ' '
  },
  {
    ' ', ' ', ':', ' ', ' ', ' ', ' ', LCDSubsystem::symbolDay,
    ' ', ' ', ' ', LCDSubsystem::symbolTime, ' ', ' ', ' ', LCDSubsystem::symbolCelcium,
    ' ', ' ', ' ', ' ', ' ', ' ', ' ', LCDSubsystem::symbolNight,
    ' ', ' ', ' ', LCDSubsystem::symbolTime, ' ', ' ', ' ', LCDSubsystem::symbolCelcium
  },
  {
    LCDSubsystem::symbolWeekday, ' ', ' ', ' ', ':', ' ', ' ', ' ',
    ' ', LCDSubsystem::symbolManualLight, ' ', ' ', ' ', ':', ' ', ' ',
    LCDSubsystem::symbolWeekend, ' ', ' ', ' ', ':', ' ', ' ', ' ',
    ' ', LCDSubsystem::symbolWatering, ' ', ' ', ' ', ' ', ' ', ' '
  },
  {
    'h', 'e', 'a', 'd', '@', 'm', 'a', 'x', 'a', 'm', 'u', 'x', 'a', '.', ' ', ' ',
    '0', '8', '-', 'a', 'p', 'r', '-', '2', '0', '1', '0', ' ', ' ', 'c', 'o', 'm'
  }
};

//common information, non editable
LCDSubsystem::FieldDescriptor LCDSubsystem::m_firstScreen[] =
{
 { fieldTime_hour, false, 0, 0, event_ClockData_changed },
 { fieldTime_minute, false, 3, 0, event_ClockData_changed },
 { fieldState, false, 0, 0, event_StateData_changed }, //this field has its own placements on the screen
 { fieldCurrentTemperature_1, false, 12, 0, event_TemperatureData_changed },
 { fieldCurrentTemperature_2, false, 12, 1, event_TemperatureData_changed },
 { fieldLastField }
};

//change time, climate
LCDSubsystem::FieldDescriptor LCDSubsystem::m_secondScreen[] =
{
 { fieldTime_hour, true, 0, 0, event_ClockData_changed, 2, 1 },
 { fieldTime_minute, true, 3, 0, event_ClockData_changed, 2, 1 },
 { fieldDayOfWeek, true, 0, 1, event_ClockData_changed, 4, 1 },
 { fieldClimateDayLength, true, 9, 0, event_ClimateData_changed, 2, 2 },
 { fieldClimateDayTemperature, true, 13, 0, event_ClimateData_changed, 2, 1 },
 { fieldClimateNightTemperature, true, 13, 1, event_ClimateData_changed, 2, 1 },
 { fieldLastField } 
};

//change dawns, manual light duration
LCDSubsystem::FieldDescriptor LCDSubsystem::m_thirdScreen[] =
{
 { fieldDawnTimeWeekday_hour, true, 2, 0, event_HeapData_changed, 2, 1 },
 { fieldDawnTimeWeekday_minute, true, 5, 0, event_HeapData_changed, 2, 1 },
 { fieldDawnTimeWeekend_hour, true, 2, 1, event_HeapData_changed, 2, 1 },
 { fieldDawnTimeWeekend_minute, true, 5, 1, event_HeapData_changed, 2, 1 },
 { fieldManualLampTime_hour, true, 11, 0, event_HeapData_changed, 2, 1 },
 { fieldManualLampTime_minute, true, 14, 0, event_HeapData_changed, 2, 1 },
 { fieldClimateWateringInterval, true, 11, 1, event_ClimateData_changed, 2, 1 },
 { fieldLastField }
};

LCDSubsystem::FieldDescriptor LCDSubsystem::m_fourthScreen[] =
{
 { fieldLastField }
};

void LCDSubsystem::refreshScreen()
{
 FieldDescriptor* fields = getCurrentFields();

 //draw the static part
 {
   uint8_t i;
   uint8_t* data = m_blankScreens[ m_screen ];

   lcd_gotoxy( 0, 0 );
   for( i = 0; i < 16; i++ )
    lcd_putc( data[ i ] );
   lcd_gotoxy( 0, 1 );
   for( ; i < 32; i++ )
    lcd_putc( data[ i ] );
 }

 uint8_t i;
 for( i = 0; fields[ i ].m_id != fieldLastField; i++ )
 {
  showField( fields[ i ] );
 }
}

LCDSubsystem::FieldDescriptor* LCDSubsystem::getCurrentFields()
{
 switch( m_screen )
 {
  case 0 : return m_firstScreen;
  case 1 : return m_secondScreen;
  case 2 : return m_thirdScreen;
  case 3 : return m_fourthScreen;
  default : return 0; //will be a crash here :)
 }
}

void LCDSubsystem::setState( States state )
{
  switch( state )
  {
  case stateDefault :
    m_state = stateDefault;
    refreshScreen();
    return;
  case stateEdit :
    {
      //find the nearest editable value
      m_cursor = 0;
      if( !findEditableDown( m_cursor ) )
        return;
      else
        m_state = stateEdit;
    }
    break;
  }
 m_state = state;
}

bool LCDSubsystem::findEditableUp( uint8_t& startValue )
{
  FieldDescriptor* fields = getCurrentFields();
  int8_t i;

  for( i = startValue; i >= 0 && !fields[ i ].m_isEditable; i-- );
  if( i >= 0 )
  {
    startValue = i;
    return true;
  }
  else
    return false;
}

bool LCDSubsystem::findEditableDown( uint8_t& startValue )
{
  FieldDescriptor* fields = getCurrentFields();
  uint8_t i;

  for( i = startValue; fields[ i ].m_id != fieldLastField && !fields[ i ].m_isEditable; i++ );
  if( fields[ i ].m_id != fieldLastField )
  {
    startValue = i;
    return true;
  }
  else
    return false;
}

void LCDSubsystem::editField( bool increase, bool repeated )
{
 uint8_t* value; //thanks the Universe, all my values are of that type
 DataStorage* dataStorage;
 uint8_t lowValue, topValue, fastPace;

 switch( getCurrentFields()[ m_cursor ].m_id )
 {
  case fieldTime_hour:
   value = &clockData.m_time.m_hour;
   dataStorage = &clockData;
   lowValue = 0;
   topValue = 23;
   fastPace = 6;
   break;
  case fieldTime_minute:
   value = &clockData.m_time.m_minute;
   dataStorage = &clockData;
   lowValue = 0;
   topValue = 59;
   fastPace = 15;
   break;
  case fieldDayOfWeek:
   value = &clockData.m_dayOfWeek;
   dataStorage = &clockData;
   lowValue = 0;
   topValue = 6;
   fastPace = 2;
   break;
  case fieldClimateDayLength:
   value = &climateData.m_dayLength;
   dataStorage = &climateData;
   lowValue = 0;
   topValue = 24;
   fastPace = 6;
   break;
  case fieldClimateDayTemperature:
   value = &climateData.m_dayTemperature;
   dataStorage = &climateData;
   lowValue = 10;
   topValue = 40;
   fastPace = 10;
   break;
  case fieldClimateNightTemperature:
   value = &climateData.m_nightTemperature;
   dataStorage = &climateData;
   lowValue = 10;
   topValue = 40;
   fastPace = 10;
   break;
  case fieldClimateWateringInterval:
   value = &climateData.m_wateringInterval;
   dataStorage = &climateData;
   lowValue = 1;
   topValue = 14;
   fastPace = 4;
   break;
  case fieldDawnTimeWeekday_hour:
   value = &heapData.m_weekdayDawn.m_hour;
   dataStorage = &heapData;
   lowValue = 0;
   topValue = 23;
   fastPace = 6;
   break;
  case fieldDawnTimeWeekday_minute:
   value = &heapData.m_weekdayDawn.m_minute;
   dataStorage = &heapData;
   lowValue = 0;
   topValue = 59;
   fastPace = 15;
   break;
  case fieldDawnTimeWeekend_hour:
   value = &heapData.m_weekendDawn.m_hour;
   dataStorage = &heapData;
   lowValue = 0;
   topValue = 23;
   fastPace = 6;
   break;
  case fieldDawnTimeWeekend_minute:
   value = &heapData.m_weekendDawn.m_minute;
   dataStorage = &heapData;
   lowValue = 0;
   topValue = 59;
   fastPace = 15;
   break;
  //note: this interval should not exceed 255 minutes
  case fieldManualLampTime_hour:
   value = &heapData.m_manualLightInterval.m_hour;
   dataStorage = &heapData;
   lowValue = 0;
   topValue = 2;
   fastPace = 2;
   break;
  case fieldManualLampTime_minute:
   value = &heapData.m_manualLightInterval.m_minute;
   dataStorage = &heapData;
   lowValue = 0;
   topValue = 59;
   fastPace = 15;
   break;
 }
 
 //universal engine to change values
 bool changed = false;
 if( increase && *value < topValue )
 {
  if( repeated )
  {
   *value += fastPace;
   if( *value > topValue )
    *value = topValue;
  }
  else
   ( *value )++;
  
  changed = true;
 }
 if( !increase && *value > lowValue )
 {
  if( repeated )
  {
   if( lowValue + fastPace <= *value )
    *value -= fastPace;
   else
    *value = 0;
  }
  else
   ( *value )--;
  
  changed = true;
 }
 
 if( changed )
  dataStorage->fireEvent_changed();
}

void LCDSubsystem::showTemperature( int8_t temperature )
{
 if( temperature < 0 )
  lcd_puts( "<0" );
 else
 {
  if( temperature > 99 )
   temperature = 99;
  lcd_puti_2d_z( temperature );
 }
}

//
// ButtonsSubsystem
//
void ButtonsSubsystem::eventsHandler( Events event )
{
 switch( event )
 {
  case event_ButtonsData_changed:
  //check the state
  {
   uint8_t i, bit = 1;
   m_pressed = m_released = m_buttons = 0;
   for( i = 0; i < 5; i++ )
   {
    int limit;

    if( buttonsData.m_buttons[ i ] > m_sensitivities[ i ] )
    {
     m_buttons |= bit;
     if( !( m_previousButtons & bit ) )
     {
      m_previousButtons |= bit;
      m_pressed |= bit;

      switch( i )
      {
       case 0: throwEvent( event_ButtonsSubsystem_button0Pressed ); break;
       case 1: throwEvent( event_ButtonsSubsystem_button1Pressed ); break;
       case 2: throwEvent( event_ButtonsSubsystem_button2Pressed ); break;
       case 3: throwEvent( event_ButtonsSubsystem_button3Pressed ); break;
       case 4: throwEvent( event_ButtonsSubsystem_button4Pressed ); break;
      }
     }
    }
    else
    {
     if( m_previousButtons & bit )
     {
      m_previousButtons ^= bit;
      m_released |= bit;
      
      switch( i )
      {
       case 0: throwEvent( event_ButtonsSubsystem_button0Released ); break;
       case 1: throwEvent( event_ButtonsSubsystem_button1Released ); break;
       case 2: throwEvent( event_ButtonsSubsystem_button2Released ); break;
       case 3: throwEvent( event_ButtonsSubsystem_button3Released ); break;
       case 4: throwEvent( event_ButtonsSubsystem_button4Released ); break;
      }
     }
    }
    bit <<= 1;
   }

   if( m_pressed || m_released )
    buttons_event();
  }
  break;
  case event_SystemTime_40ms:
   m_idleTime++;
   if( m_idleTime == 30 )
   {
    m_idleTime = 0;
    timeout_event();
   }
   break;
 }
}

void ButtonsSubsystem::buttons_event()
{
 //event_ButtonsData_changed event is coming 25 times a second, only the filtered 
 //event 'buttons are really changed' should drop this value
 m_idleTime = 0;

 if( m_state & stateWave )
  if( checkWave() )
   return; //this event was used to finish the wave

 //global change
 if( ( m_state & stateUser ) && ( m_released & 0x03 ) )
 {
  m_state = 0;
  throwEvent( event_ButtonsSubsystem_userLeaves );
  return;
 }
 
 switch( m_state )
 {
  case 0:
   switch( m_pressed )
   {
    case 0x02:
     startWave( waveLightLamp );
     return;
    case 0x01:
     startWave( wavePutOffLamp );
     return;
   }
   return;
  case stateUser:
   switch( m_pressed )
   {
    case 0x04:
     startWave( waveShiftDown );
     return;
    case 0x10:
     startWave( waveShiftUp );
     return;
   }
   return;
  case stateUser | stateInput:
   if( m_released & 0x08 )
   {
    m_state = stateUser;
    throwEvent( event_ButtonsSubsystem_stopInput );
    return;
   }
   switch( m_pressed )
   {
    case 0x10:
     throwEvent( event_ButtonsSubsystem_inputUp );
     return;
    case 0x04:
     throwEvent( event_ButtonsSubsystem_inputDown );
     return;
   }
   return;
  case stateUser | stateSwitchingMode:
   //this mode is active before all 3 its buttons are released
   if( !( m_buttons & 0x1C ) )
   {
    m_state &= ~stateSwitchingMode;
   }
   return;
 }
}

void ButtonsSubsystem::timeout_event()
{
 //any wave is stopped if the timeout occurs
 m_state &= ~stateWave;
 
 switch( m_state )
 {
  case 0:
   switch( m_buttons )
   {
    case 0x03: //both rear buttons
    {
     m_state = stateUser;
     throwEvent( event_ButtonsSubsystem_userComes );
    }
    return;
    case 0x1C: //all front buttons
     throwEvent( event_ButtonsSubsystem_watering );
    return;
   }
   return;
  case stateUser:
   if( ( m_buttons & 0x1C ) == 0x1C )
   {
    m_state = stateUser | stateSwitchingMode;
    throwEvent( event_ButtonsSubsystem_switchMode );
    return;
   }
   if( ( m_buttons & 0x1C ) == 0x08 )
   {
    m_state = stateUser | stateInput;
    throwEvent( event_ButtonsSubsystem_startInput );
   }
   return;
  case stateUser | stateInput:
   switch( m_buttons & 0x1C )
   {
    case 0x18 :
     throwEvent( event_ButtonsSubsystem_inputRepeatedUp );
     return;
    case 0x0C:
     throwEvent( event_ButtonsSubsystem_inputRepeatedDown );
     return;
   }
 }
}

void ButtonsSubsystem::startWave( WaveTypes type )
{
 m_state |= stateWave;
 m_wave.m_pressIndex = m_wave.m_releaseIndex = 0;
 m_wave.m_type = type;

 //initial shift
 checkWave();
}

bool ButtonsSubsystem::checkWave()
{
 bool notBroken = true;
 uint8_t const*const& descriptor = m_waveDescriptors[ m_wave.m_type ];

 if( m_pressed )
 {
  if( descriptor[ m_wave.m_pressIndex ] == m_pressed )
   m_wave.m_pressIndex++;
  else
   notBroken = false;
 }
 if( m_released && notBroken )
 {
  if( descriptor[ m_wave.m_releaseIndex ] == m_released )
   m_wave.m_releaseIndex++;
  else
   notBroken = false;
 }
 if( notBroken && ( m_wave.m_releaseIndex > m_wave.m_pressIndex ) )
  notBroken = false;

 if( notBroken && !descriptor[ m_wave.m_releaseIndex ] )
 {
  switch( m_wave.m_type )
  {
   case waveLightLamp:
    throwEvent( event_ButtonsSubsystem_lightLamps );
    break;
   case wavePutOffLamp:
    throwEvent( event_ButtonsSubsystem_putOutLamps );
    break;
   case waveShiftUp:
    throwEvent( event_ButtonsSubsystem_shiftUp );
    break;
   case waveShiftDown:
    throwEvent( event_ButtonsSubsystem_shiftDown );
    break;
  }
  
  m_state &= ~stateWave;
  return true;
 }
  
 if( !notBroken )
 {
  m_state &= ~stateWave;
 }
 
 return false;
}

uint8_t ButtonsSubsystem::m_waveDescriptors[ 4 ][ 4 ] =
{
 { 2, 1, 0 }, //light the lamp
 { 1, 2, 0 }, //disable the lamp
 { 16, 8, 4, 0 }, //shift up
 { 4, 8, 16, 0 } //shift down
};

//I was tired of seeing my buttons with different sensitivities,
//so make different edges for their signals.
uint8_t ButtonsSubsystem::m_sensitivities[ 5 ] = { 20, 55, 25, 25, 35 };

//
// LEDSubsystem
//
LEDSubsystem::LEDSubsystem()
 : m_wateringValue( 0 ), m_heartCurrent( 0 ), m_heartTop( 0 )
{
 uint8_t i;
 for( i = 0; i < ledNumberOfLeds; i++ )
 {
  State& led = m_leds[ i ];
  led.m_value = 0;
  led.m_order.m_type = orderNone; 
  setValue( i, 0 );
 }
}

//not very good solution, bug it is 2:13 after the midnight, I am tired :)
#include "twi.h"
extern TWI Twi;
         
void LEDSubsystem::eventsHandler( Events event )
{
 //it can be some button
 {
  uint8_t index = 10;
  bool released = false;
  switch( event )
  {
   case event_ButtonsSubsystem_button0Released:
    released = true;
   case event_ButtonsSubsystem_button0Pressed:
    index = ledButton0; break;
   case event_ButtonsSubsystem_button1Released:
    released = true;
   case event_ButtonsSubsystem_button1Pressed:
    index = ledButton1; break;
   case event_ButtonsSubsystem_button2Released:
    released = true;
   case event_ButtonsSubsystem_button2Pressed:
    index = ledButton2; break;
   case event_ButtonsSubsystem_button3Released:
    released = true;
   case event_ButtonsSubsystem_button3Pressed:
    index = ledButton3; break;
   case event_ButtonsSubsystem_button4Released:
    released = true;
   case event_ButtonsSubsystem_button4Pressed:
    index = ledButton4; break;
  }

  if( index != 10 )
  {
   Order& order = m_leds[ index ].m_order;
   if( released )
   {
    //fade out
    order.m_type = orderNone;
   }
   else
   {
    //light
    order.m_value = 99;
    order.m_type = orderSet;
   }
   return;
  }
 }
 
 //light/put off the backlight
 switch( event )
 {
  case event_ButtonsSubsystem_userComes:
   m_leds[ ledLCDBacklight ].m_order.m_type = orderSet;
   m_leds[ ledLCDBacklight ].m_order.m_value = 99;
   m_leds[ ledLCDBacklight ].m_order.m_timeBefore = 0;
   return;
  case event_ButtonsSubsystem_userLeaves:
   m_leds[ ledLCDBacklight ].m_order.m_type = orderNone;
   m_leds[ ledLCDBacklight ].m_order.m_timeBefore = 90;
   return;
 }
 
 if( event != event_SystemTime_40ms )
  return;

 //update the state leds
 updateStateLeds();
 
 //compute all leds
 uint8_t index;
 for( index = 0; index < ledNumberOfLeds; index++ )
 {
  State& led = m_leds[ index ];
  Order& order = led.m_order;
  uint8_t oldValue = led.m_value;

  //timer
  if( order.m_timeBefore )
  {
   order.m_timeBefore--;
   continue;
  }
  
  //some rule is active
  if( order.m_type != orderNone )
  {
   //go up
   if( led.m_value < order.m_value )
   {
    if( led.m_value + 40 <= order.m_value )
     led.m_value += 40;
    else
     led.m_value = order.m_value;
   }
   else
   {
    //go down
    if( led.m_value > order.m_value )
    {
     if( led.m_value >= order.m_value + 40 )
      led.m_value -= 40;
     else
      led.m_value = order.m_value;
    }
    else
     //stop the peak (if defined)
     if( order.m_type == orderPeak )
      order.m_type = orderNone;
   }
  }
  else
  {
   //else, just decrease the value
   if( led.m_value > 3 )
    led.m_value -= 3;
   else
    led.m_value = 0;
  }

  if( led.m_value != oldValue )
   setValue( index, led.m_value );
 }
}

//the smooth fuction based on 'exp'
static uint8_t smoothFunction[ 100 ] =
{
 0       , 0       , 0       , 0       , 0       , 0       , 0       , 0       ,
 0       , 0       , 0       , 0       , 0       , 0       , 0       , 0       ,
 0       , 0       , 0       , 0       , 0       , 0       , 0       , 0       ,
 0       , 0       , 0       , 0       , 0       , 0       , 0       , 0       ,
 0       , 0       , 0       , 0       , 0       , 0       , 0       , 0       ,
 0       , 0       , 0       , 0       , 0       , 0       , 0       , 0       ,
 0       , 0       , 0       , 0       , 1       , 1       , 1       , 1       ,
 1       , 1       , 1       , 2       , 2       , 2       , 2       , 3       ,
 3       , 3       , 4       , 4       , 5       , 5       , 6       , 7       ,
 7       , 8       , 9       , 10      , 11      , 12      , 14      , 15      ,
 17      , 19      , 21      , 23      , 25      , 28      , 31      , 34      ,
 38      , 42      , 47      , 52      , 57      , 63      , 70      , 77      ,
 85      , 94      , 104     , 115
};

void LEDSubsystem::setValue( uint8_t ledIndex, uint8_t value )
{
 uint8_t address;

 switch( ledIndex )
 {
  case ledButton0: address = 0x80; break;
  case ledButton1: address = 0x81; break;
  case ledButton2: address = 0x82; break;
  case ledButton3: address = 0x83; break;
  case ledButton4: address = 0x84; break;
  case ledLCDBacklight: address = 0x85; break;
  case ledHeart: address = 0x86; break;
  case ledWatering: address = 0x87; break;
  default: return;
 }

 Twi.sendData( address, smoothFunction[ value ] );
}

void LEDSubsystem::updateStateLeds()
{
 Order& heartOrder = m_leds[ ledHeart ].m_order;
 Order& wateringOrder = m_leds[ ledWatering ].m_order;

 //heart
 if( m_heartCurrent++ == m_heartTop )
 {
  m_heartCurrent = 0;
  
  //calculate new value
  m_heartTop = 50; //still
  if( stateData.m_watering )
   m_heartTop -=15;
  if( stateData.m_badTemperature )
   m_heartTop -= 15;
  
  //light the led
  heartOrder.m_type = orderPeak;
  heartOrder.m_value = 90;
 }
 //the second bit
 if( m_heartCurrent == 6 )
 {
  //light the led
  heartOrder.m_type = orderPeak;
  heartOrder.m_value = 99;
 }
 
 //watering
 if( stateData.m_watering )
 {
  if( ++m_wateringValue == 6 )
  {
   m_wateringValue = 0;
   wateringOrder.m_type = orderPeak;
   wateringOrder.m_value = 80;
  }
 }
}

LEDSubsystem::Order::Order()
 : m_value( 0 ), m_type( orderNone ), m_timeBefore( 0 )
{
}

//
// DaySubsystem
//
void DaySubsystem::eventsHandler( Events event )
{
  //the only reaction to any event is that calculation of days/nights,
  //so we do not check type of the event

  //get the dawn & the sunset
  Time 
    dawn = clockData.isWeekday() ? heapData.m_weekdayDawn : heapData.m_weekendDawn,
    sunset = dawn;
  sunset.m_hour += climateData.m_dayLength; //it can exceed 24

  //very easy kind of the Divine Providence
  bool isDayNow = clockData.m_time > dawn && sunset > clockData.m_time;
  if( stateData.m_dayNow != isDayNow )
  {
    stateData.m_dayNow = isDayNow;
    stateData.fireEvent_changed();
  }
}

//
// LampSubsystem
//
LampSubsystem::LampSubsystem()
: m_timeout( 0 ), m_isDay( false )
{
 //this pin controls the lamps
 DDRC |= _BV( PIN4 );
}

void LampSubsystem::eventsHandler( Events event )
{
  switch( event )
  {
  case event_ClockSubsystem_minuteElapsed:
    //it effects on the manual light
    if( m_timeout )
    {
      m_timeout--;
      if( !m_timeout )
        updateToNormalState();
    }
    return;
  case event_ButtonsSubsystem_lightLamps:
    if( m_isDay )
    {
     if( m_timeout ) //it means the cancel
      updateToNormalState();
    }
    else
      setInvertState();
    return;
  case event_ButtonsSubsystem_putOutLamps:
    if( !m_isDay )
    {
     if( m_timeout ) //it means the cancel
      updateToNormalState();
    }
    else
     setInvertState();
    return;
  case event_StateData_changed:
    if( m_isDay != stateData.m_dayNow )
    {
      m_isDay = stateData.m_dayNow;
      updateToNormalState();
    }
    return;
  }
}

void LampSubsystem::updateToNormalState()
{
  m_timeout = 0;
  if( m_isDay )
    PORTC |= _BV( PIN4 );
  else
    PORTC &= ~_BV( PIN4 );
}

void LampSubsystem::setInvertState()
{
 m_timeout = heapData.m_manualLightInterval.m_hour * 60 + heapData.m_manualLightInterval.m_minute;
 if( m_isDay )
  PORTC &= ~_BV( PIN4 );
 else
  PORTC |= _BV( PIN4 );
}

//
// HeaterSubsystem
//
HeaterSubsystem::HeaterSubsystem()
 : m_globalLoop( 0 )
{
 m_idleTime[ 0 ] = m_idleTime[ 1 ] = 0;
 
 //these pins control the heaters
 DDRC |= _BV( PIN5 ) | _BV( PIN6 );
}

void HeaterSubsystem::eventsHandler( Events event )
{
  switch( event )
  {
  case event_ClockSubsystem_minuteElapsed:
    {
      uint8_t i;
      bool error = false;
      for( i = 0; i < 2; i++ )
      {
        if( m_idleTime[ i ] < 120 ) //so, there is that limit for the idle states
          m_idleTime[ i ]++;
        else
         error = true;
      }

      //set/drop a flag
      if( error != stateData.m_badTemperature )
      {
       stateData.m_badTemperature = error;
       stateData.fireEvent_changed();
      }
    }
    break;
   case event_SystemTime_second:
   {
    m_globalLoop++;

    //drop the loop and check if a heater should be enabled
    if( m_globalLoop == 60 )
    {
     m_globalLoop = 0;

     uint8_t i;
     bool newValue;
     uint8_t
       optimalTemperature = stateData.m_dayNow ? climateData.m_dayTemperature : climateData.m_nightTemperature;
     int8_t currentTemperature;

     for( i = 0; i < 2; i++ )
     {
      //formally, the temperature could be less than zero, possible signed/unsigned mismatch
      currentTemperature = temperatureData.m_temperature[ i ];
      if( currentTemperature < 0 )
       currentTemperature = 0;

      newValue = currentTemperature < optimalTemperature;

      //drop the idle time if the state is changed, change the indicator
      if( newValue != stateData.m_heaterWorks[ i ] )
      {
       m_idleTime[ i ] = 0;
       stateData.m_heaterWorks[ i ] = newValue;
       stateData.fireEvent_changed();
      }

      //use the value
      uint8_t mask = i ? _BV( PIN6 ) : _BV( PIN5 );
      if( newValue )
       PORTC |= mask;
      else
       PORTC &= ~mask;
     }
    }
   }
   break;
  }
}

//
// WateringSubsystem
//
WateringSubsystem::WateringSubsystem()
 : m_dayCounter( 0 )
{
}

void WateringSubsystem::eventsHandler( Events event )
{
 //waiting for the watering
 if( stateData.m_watering )
 {
  if( event == event_ButtonsSubsystem_watering )
  {
   stateData.m_watering = false;
   stateData.fireEvent_changed();
  }
 }
 //check if we should set the flag
 else
 {
  if( event == event_ClockSubsystem_hourElapsed )
  {
   Time& time = clockData.m_time;
   //do it once a day
   if( time.m_hour == 18 && !time.m_minute && !time.m_second )
   {
    if( m_dayCounter < climateData.m_wateringInterval )
      m_dayCounter++;
    else
    {
     m_dayCounter = 0;
     stateData.m_watering = true;
     stateData.fireEvent_changed();
    }
   }
  }
 }
}

//the only objects used everywhere
ClockSubsystem clockSubsystem;
LCDSubsystem lcdSubsystem;
ButtonsSubsystem buttonsSubsystem;
LEDSubsystem ledSubsystem;
DaySubsystem daySubsystem;
LampSubsystem lampSubsystem;
HeaterSubsystem heaterSubsystem;
WateringSubsystem wateringSubsystem;