/*
 * Project: Device in Jar. Greenhouse.
 * Device:  Main Atmega32
 * File:    One of the bases of the idea. Subsystems represent small 'individual' beings of the system that
 *          receive some events, think something, change some data and send other events
 *
 * License: freeware. Impressed by other cool ideas, will be happy if I inspire to new ideas.
 *
 * © 2010 by Maxim Ushakov; www.maxamuxa.com; head@maxamuxa.com
 */
#ifndef __SUBSYSTEM_H__
#define __SUBSYSTEM_H__

#include <avr/io.h>

#include "events.h"

class Subsystem
{
 public:
  //This method is called when any subscribed event occurs.
  virtual void eventsHandler( Events event ) = 0;
};

//it receives ticks from the timer and supports the system time
class ClockSubsystem : public Subsystem
{
 public:
  virtual void eventsHandler( Events event );
};

//The interface with user:
//it displays all possible information, pages screens, processes editing.
class LCDSubsystem : public Subsystem
{
 public:
  LCDSubsystem();
  
  virtual void eventsHandler( Events event );
  void initialize();

  enum CustomSymbols
  {
   symbolWatering,
   symbolManualLight,
   symbolCelcium,
   symbolDay,
   symbolNight,
   symbolWeekday,
   symbolWeekend,
   symbolTime
  };
  
 private:
  enum States
  {
   stateDefault, //display time, etc
   stateEdit //edit mode is selected, so cursor should exist
  };
  
  //elementary data fields
  enum Fields
  {
   //time
   fieldTime_hour,
   fieldTime_minute,
   fieldDayOfWeek,
   
   //current state and around that (many fields around the screen)
   fieldState,
  
   //current temperatures
   fieldCurrentTemperature_1,
   fieldCurrentTemperature_2,
   
   //setting for the climate
   fieldClimateDayLength,
   fieldClimateDayTemperature,
   fieldClimateNightTemperature,
   fieldClimateWateringInterval,
   
   //morning alarms
   fieldDawnTimeWeekday_hour,
   fieldDawnTimeWeekday_minute,
   fieldDawnTimeWeekend_hour,
   fieldDawnTimeWeekend_minute,
   
   //time delay for the manual lamp
   fieldManualLampTime_hour,
   fieldManualLampTime_minute,
   
   //to mark ends of field lists
   fieldLastField
  };

  struct FieldDescriptor
  {
   Fields m_id;
   bool m_isEditable;
   uint8_t m_x, m_y;
   
   //Each field follows some data storage.
   //It is handy to hold its event to know when should we update the field.
   Events m_eventToUpdate;

   //The rectangle to clear (the fields blink when are being edited)
   uint8_t m_width, m_height;
  };
  
  //show a field according to the real settings
  void showField( FieldDescriptor& field );

  //redraw the current screen
  void refreshScreen();
  
  FieldDescriptor* getCurrentFields();
  
  void setState( States state );

  //Try to find the nearest down/up editable values in the current screen.
  //Return false if not found.
  bool findEditableDown( uint8_t& startValue );
  bool findEditableUp( uint8_t& startValue );

  //Edit the current field.
  //1. a value is rising if increase == true (falling elsewise)
  //2. repeated means a fast change due to the holding button
  //3. this method cares about ranges for each known value
  void editField( bool increase, bool repeated );
  
  void showTemperature( int8_t temperature );
  
  //rasters for custom symbols
  static uint8_t m_customSymbols[ 64 ];

  //blank screens with all static symbols
  static uint8_t m_blankScreens[ 4 ][ 32 ];

  //these arrays define all screens
  static FieldDescriptor
    m_firstScreen[],
    m_secondScreen[],
    m_thirdScreen[],
    m_fourthScreen[];

  //current state
  States m_state;
  uint8_t m_screen;
  uint8_t m_cursor;
};


//This subsystem controls incoming changes found by the ADC (it tracks buttons) and 
//post high-level events.
class ButtonsSubsystem : public Subsystem
{
 public:
  virtual void eventsHandler( Events event );
  
 private:
  void buttons_event();
  void timeout_event();

  enum WaveTypes
  {
    waveLightLamp = 0,
    wavePutOffLamp,
    waveShiftUp,
    waveShiftDown
  };

  void startWave( WaveTypes type );
  bool checkWave(); //returns true if a wave is finished successfully
  
  enum States
  {
   stateUser = 0x01, //user takes the bottle
   stateInput = 0x02, //user pressed the middle button to change values
   stateWave = 0x04, //some wave is in progress
   stateSwitchingMode = 0x08 //when user holds 3 buttons, we should what until he release them
  };
  
  struct Wave
  {
   WaveTypes m_type;
   uint8_t m_pressIndex, m_releaseIndex;
  } m_wave;
  
  //bits for buttons:
  // 0 - back bottom
  // 1 - back top
  // 2 - front bottom
  // 3 - front middle
  // 4 - front top
  
  uint8_t m_buttons; //current pressed buttons
  uint8_t m_previousButtons; //mask of pressed buttons before
  uint8_t m_pressed, m_released; //masks for buttons (current events)
  uint8_t m_idleTime; //is used to generate the timeout event
  uint8_t m_state; //mask
  
  static uint8_t m_waveDescriptors[ 4 ][ 4 ];
  static uint8_t m_sensitivities[ 5 ];
};

//It traces different pieces of the state and operate with LEDs (backlight of buttons, 'heart', etc).
class LEDSubsystem : public Subsystem
{
 public:
  LEDSubsystem();
  virtual void eventsHandler( Events event );
  
 private:
  enum LEDIndices
  {
   ledButton0 = 0,
   ledButton1,
   ledButton2,
   ledButton3,
   ledButton4,
   ledLCDBacklight,
   ledHeart,
   ledWatering,
   ledNumberOfLeds
  };

  enum OrderType
  {
   orderPeak, //to reach the value and go to 'orderNone' state to fade off
   orderSet, //reach the value and hold that
   orderNone //no order, slow fading off
  };

  struct Order
  {
   Order();

   uint8_t m_value;
   OrderType m_type;
   //the rule is not applied before this timer is positive (40ms for each one).
   //When no rule is defined, the value is hold constant.
   uint8_t m_timeBefore;
  };
  
  struct State
  {
   Order m_order;
   uint8_t m_value;
  };
  
  //Sends some value via TWI.
  //Note: 
  // 1. it uses some smoothing math function
  // 2. the maximal value is 127
  void setValue( uint8_t ledIndex, uint8_t value );

  //it is called 25 times per second, operates with watering, heart leds, etc
  void updateStateLeds();
  
  //the collection :)
  State m_leds[ ledNumberOfLeds ];
  
  //special value for the watering
  uint8_t m_wateringValue;
  
  //values for the heart
  uint8_t m_heartCurrent, m_heartTop;
};

//This subsystem just calculates if it is day or night (based on other parameters).
class DaySubsystem : public Subsystem
{
public:
  virtual void eventsHandler( Events event );
};

//It light or put out the light according to the settings (or if set manually).
class LampSubsystem : public Subsystem
{
public:
  LampSubsystem();
  virtual void eventsHandler( Events event );

private:
  void updateToNormalState();
  void setInvertState();

  //if not zero, then the minutes before the manual light is put out
  uint8_t m_timeout;

  //the last read value from the stateData (to understand if we should cancel the manual light)
  bool m_isDay;
};

//This subsystem watches the temperatures and switch the heaters on/off.
//Also, it determines if a temperature is too big/small.
class HeaterSubsystem : public Subsystem
{
public:
  HeaterSubsystem();
  virtual void eventsHandler( Events event );

private:
  //it defines either on & off states (seconds from beginning of the current loop)
  uint8_t m_globalLoop;

  //minutes since a state of heater (on/off) was changed
  uint8_t m_idleTime[ 2 ];

  //this flag increases each minute and drops if changes happened to a heater
  bool m_heaterChanged[ 2 ];
};

//This one measures time to watering and show its alert
class WateringSubsystem : public Subsystem
{
 public:
  WateringSubsystem();
  virtual void eventsHandler( Events event );
  
 private:
  uint8_t m_dayCounter;
};

//
// The only data storages with global scope
//
extern ClockSubsystem clockSubsystem;
extern LCDSubsystem lcdSubsystem;
extern ButtonsSubsystem buttonsSubsystem;
extern LEDSubsystem ledSubsystem;
extern DaySubsystem daySubsystem;
extern LampSubsystem lampSubsystem;
extern HeaterSubsystem heaterSubsystem;
extern WateringSubsystem wateringSubsystem;

#endif //__SUBSYSTEM_H__