 /*
 * Project: Device in Jar. Greenhouse.
 * Device:  Main Atmega32
 * File:    TWI for the master Atmega32 (transmit/receive 16-bit words)
 *
 * License: freeware. Impressed by other cool ideas, will be happy if I inspire to new ideas.
 *
 * © 2010 by Maxim Ushakov; www.maxamuxa.com; head@maxamuxa.com
 */
#include "MPUSettings.h"
#include "twi.h"
#include <util/delay.h>

#define F_TWI 100000UL //approximate speed for TWI

void TWI::enable()
{
  disable(); //to be sure of refreshing everything

  //some operations
 TWBR = uint8_t( F_CPU / F_TWI / 2 - 8 ); //speed
 TWAR = 2; //its address

 //enable internal pull-up resistors
 DDRC &= 0b11111100;
 PORTC |= 0b00000011;

 //enable pins SDA/SCL
 TWCR = _BV( TWEN );
}

void TWI::disable()
{
 TWCR = 0;
}

TWI::Result TWI::sendData( uint8_t address, uint16_t data )
{
 m_loopsCounter = 0;

 //start session
 TWCR = _BV( TWINT ) | _BV( TWSTA ) | _BV( TWEN );
 if( waitForTWINT() != Nothing )
 {
  dropState();
  return Error;
 }

 //send SLA + W
 TWDR = 4 | TW_WRITE;
 TWCR = _BV( TWINT ) | _BV( TWEN );
 if( waitForTWINT() != Nothing || TWSR != TW_MT_SLA_ACK )
 {
  dropState();
  return Error;
 }

 //send the address & 2 bytes of data
 uint8_t bytesTransmitted;
 for( bytesTransmitted = 0; bytesTransmitted < 3; bytesTransmitted++ )
 {
  switch( bytesTransmitted )
  {
   case 0 : TWDR = address; break;
   case 1 : TWDR = uint8_t( data ); break;
   case 2 : TWDR = uint8_t( data >> 8 ); break;
  }
  TWCR = _BV( TWINT ) | _BV( TWEN );
  if( waitForTWINT() != Nothing || TWSR != TW_MT_DATA_ACK )
  {
   dropState();
   return Error;
  }
 }

 //stop the session
 TWCR = _BV( TWEN ) | _BV( TWSTO ) | _BV( TWINT );
 _delay_us( 64 ); //experimental stable value for multiple writings

 return DataSent;
}

TWI::Result TWI::receiveData( uint8_t address, uint16_t& data )
{
 //zero address keeps the address where the data is given from
 Result result = sendData( 0, address );
 if( result != DataSent )
 {
  return result;
 }

 m_loopsCounter = 0;

 //start session
 TWCR = _BV( TWINT ) | _BV( TWSTA ) | _BV( TWEN );
 if( waitForTWINT() != Nothing )
 {
  dropState();
  return Error;
 }

 //send SLA + R
 TWDR = 4 | TW_READ;
 TWCR = _BV( TWINT ) | _BV( TWEN );
 if( waitForTWINT() != Nothing || TWSR != TW_MR_SLA_ACK )
 {
  dropState();
  return Error;
 }
 TWCR = _BV( TWINT ) | _BV( TWEN ) | _BV( TWEA );

 //receive the data
 uint8_t bytesReceived;
 data = 0;
 for( bytesReceived = 0; bytesReceived < 2; bytesReceived++ )
 {
  if( waitForTWINT() != Nothing )
  {
   dropState();
   return Error;
  }

  if( !bytesReceived )
  {
   data = TWDR;
   //the second byte is the last, NACK should be sent
   TWCR = _BV( TWINT ) | _BV( TWEN );
  }
  else
   data |= uint16_t( TWDR ) << 8;
 }

 //send STOP
 TWCR = _BV( TWINT ) | _BV( TWEN ) | _BV( TWSTO );

 return DataReceived;
}

TWI::Result TWI::waitForTWINT()
{
 //We define the limit here.
 //I do not investigate this issue deeply, but the value here works well
 //and it is enought to keep it.
 for( ; !( TWCR & _BV( TWINT ) ) && m_loopsCounter < 1000; m_loopsCounter++ );
 return m_loopsCounter == 1000 ? Timeout : Nothing;
}

void TWI::dropState()
{
 TWCR = _BV( TWEN ) | _BV( TWSTO );
}
